// Use the Stellent Content Access APIs to extract properties or text from documents
// Based on the Stellent cademo example

#define WINDOWS
//#define WIN32
#define FALSE 0
#define TRUE  1
#define EXTYPE_INVALID 10
#define EXTYPE_TEXT 11
#define EXTYPE_PROPERTIES 12

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include "sccca.h"

#define DELETE_FILE( filename ) DeleteFile( (filename) )
#define CLOSE_HANDLE( handle ) CloseHandle( handle )
#define WRITE_FILE( handle, buffer, buflen, outcount) \
        WriteFile( (handle), (buffer), ((DWORD)buflen), (outcount), NULL )

// Function Prototypes
void processFile(void);
void SkipTagLocal(VTHCONTENT hContent, VTDWORD dwTagType, PSCCCAGETCONTENT pcaContent);
void AddDocumentProperty(VTHCONTENT hContent, VTDWORD dwPropertyId, PSCCCAGETCONTENT pcaContent, HANDLE *TempPropFile);
void doCAfile(HANDLE *outFile);
DAERR doCA(HANDLE *outFile, VTHDOC hLocDoc);
void WriteSeparator(HANDLE *hFile);

// Property Names
char *props[] =
 {
   "Document Comment",    /*      0       */
   "Keyword",
   "Last Saved By",
   "Primary Author",
   "Subject",
   "Title",
   "Abstract",
   "Account",
   "Address",
   "Attachments",
   "Authorization",       /*      10      */
   "Backup Date",
   "Bill To",
   "Blind Copy",
   "Carbon Copy",
   "Category",
   "Checked By",
   "Client",
   "Completed Date",
   "Character Count",
   "Page Count",          /*      20      */
   "Word Count",
   "Creation Date",
   "Department",
   "Destination",
   "Disposition",
   "Division",
   "Minutes Edited",
   "Editor",
   "Forward To",
   "Group",               /*      30      */
   "Language",
   "Last Print Date",
   "Mail Stop",
   "Matter",
   "Office",
   "Operator",
   "Owner",
   "Project",
   "Publisher",
   "Purpose",             /*      40      */
   "Received From",
   "Recorded By",
   "Recorded Date",
   "Reference",
   "Revision Date",
   "Revision Notes",
   "Revision Number",
   "Secondary Author",
   "Section",
   "Security",            /*      50      */
   "Source",
   "Status",
   "Document Type",
   "Typist",
   "Version Date",
   "Version Notes",
   "Version Number",
   "Base File Location",
   "Manager",
   "Company",             /*      60      */
   "Last Saved Date",
   "Presentation Format",
   "Count Bytes",
   "Count Lines",
   "Count Paragraphs",
   "Count Slides",
   "Count Slides Hidden",
   "Count Notes",
   "Count MM Clips",
   "Count Characters with Spaces",  /*  70  */
   "Scale Crop",
   "Links Dirty",
   "Title of Parts",
   "Heading Pairs",
   "Document Number",
   "User Defined Property"            /* 76 */
 };


// Globals
VTDWORD       dwCharacterSet = SO_UNICODE; /* current character set */
VTDWORD       dwFileFormat = FI_UNICODE;   /* current file format   */
VTHDOC        hDoc = VTHDOC_INVALID;
VTDWORD       MAXDABUFSIZE = 0x1000;
char          contentFilename[1024];         /* the content file */
char          outputFilename[1024];          /* the output file */

int           nExpandAttachments = FALSE; /*  Do we expand attachments?  */
int           nExpandEmbeddings = FALSE;  /*  Do we expand embeddings?  */
int           nExpandArchives = FALSE;    /*  Do we expand archives?  */
int           nGeneratedText = FALSE;     /*  Do we process generated text?  */

// What do we extract out?
int           exType = EXTYPE_INVALID;


////////////////////////////////////////////////////////////////////////////////
main(int argc, char *argv[])
{
  if(argc == 4) {
    // Get the extraction type
    if (strcmp(argv[1], "properties") == 0) {
      exType = EXTYPE_PROPERTIES;
    }
    if (strcmp(argv[1], "text") == 0) {
      exType = EXTYPE_TEXT;
    }

    if (EXTYPE_INVALID != exType) {

      // Copy the filenames
      strncpy(contentFilename, argv[2], 1023);
      strncpy(outputFilename, argv[3], 1023);

      // Process the given document
      processFile();

      // Done
      return (0);
    }
  }

  // Print Usage
  printf("Usage: docFilter properties|text inputFilename outputFilename\n");
  return (1);
}


////////////////////////////////////////////////////////////////////////////////
void processFile(void)
{
  HANDLE *outFile;
  DAERR err;
  //WORD wBOM;
  //DWORD dwBytesWritten;

  // Initialize the data access
  err = DAInit();
  if(err == DAERR_OK) {

    // Open output file
    outFile = CreateFile( outputFilename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if (NULL == outFile) {
      perror("docFilter: cannot open output file for writing");
      return;
    }

    // If source format is UNICODE, output the UNICODE BOM (byte order mark) to text file.
    //if(FI_UNICODE == dwFileFormat) {
    //  wBOM = 0xFEFF;
    //  WRITE_FILE( outFile, (LPBYTE)&wBOM, 2, &dwBytesWritten );
    //}

    // Run the content access
    doCAfile(outFile);

    // Close the output file
    CLOSE_HANDLE(outFile);
  }

  // DeInitialize the data access
  err = DADeInit();
}


////////////////////////////////////////////////////////////////////////////////
void doCAfile(HANDLE *outFile)
{
  DAERR err;

  // Open the source file using Data Access
  //err = DAOpenDocument( &hDoc, ( VTDWORD ) IOTYPE_UNIXPATH, ( VTLPVOID ) contentFilename, 0 );
  err = DAOpenDocument( &hDoc, ( VTDWORD ) IOTYPE_ANSIPATH, ( VTLPVOID ) contentFilename, 0 );
  if (err == DAERR_OK ) {

    // Use the output character set specified by the global dwCharacterSet
    // and pen the document for the content access.
    err = DASetOption( hDoc, SCCOPT_OUTPUTCHARACTERSET, ( VTLPVOID ) & dwCharacterSet, sizeof ( VTDWORD ) );
    if ( err != DAERR_OK ) {
      DACloseDocument( hDoc );
      hDoc = VTHDOC_INVALID;
    }

    // Do the content access
    err = doCA(outFile, hDoc);

    if (err != DAERR_OK && err != DAERR_EOF) {
      // Some error occured while reading the file
      DACloseDocument( hDoc );
      hDoc = VTHDOC_INVALID;
      //fprintf(outFile, "An unknown error occurred when trying to read document.\n");
      printf("An unknown error occurred when trying to read the document.\n");
    }
  }
}


////////////////////////////////////////////////////////////////////////////////
DAERR doCA(HANDLE  *outFile, VTHDOC hLocDoc)
{
  VTHCONTENT hContent;  /* Content access content handle */
  char *hBuffer = 0;
  SCCCAGETCONTENT caContent;
  int err;
  VTDWORD dwOption = (nGeneratedText) ? SCCEX_IND_GENERATED : 0;

  DASetOption(hLocDoc, SCCOPT_CA_FLAGS, (VTLPVOID)&dwOption, sizeof(VTDWORD));

  // Open the document for content access
  CAOpenContent( hLocDoc, (VTLPHCONTENT)&hContent );

  // Initialize SCCCAGETCONTENT structure 
  hBuffer = (char *) malloc( MAXDABUFSIZE );
  *hBuffer = 0;
  caContent.dwStructSize = sizeof ( SCCCAGETCONTENT );
  caContent.dwFlags = 0;
  caContent.dwMaxBufSize = MAXDABUFSIZE;
  caContent.pDataBuf = ( char * ) ( hBuffer );

  // Use Content Access to read the text from the file as unicode
  err = CAReadFirst( hContent, &caContent );
  if ( err != DAERR_OK ) {
    return err;
  }

  while ( err == DAERR_OK ) {
    DWORD dwBytesWritten;
    switch ( caContent.dwType ) {
    case SCCCA_SHEET:

      if (EXTYPE_TEXT == exType) {
	/*  Section information.  For now it only contains the name.  */
	WRITE_FILE(outFile, caContent.pDataBuf, caContent.dwDataBufSize, &dwBytesWritten);
	WriteSeparator(outFile);
      }
      break;

    case SCCCA_GENERATED:
      if (nGeneratedText != FALSE) {
	if (EXTYPE_TEXT == exType) {
	  WRITE_FILE(outFile, caContent.pDataBuf, caContent.dwDataBufSize, &dwBytesWritten);
	  WriteSeparator(outFile);
	}
      }
      break;

    case SCCCA_TEXT:
      if (EXTYPE_TEXT == exType) {
	WRITE_FILE(outFile, caContent.pDataBuf, caContent.dwDataBufSize, &dwBytesWritten);
	WriteSeparator(outFile);
      }
      break;

    case SCCCA_BEGINTAG:
      switch ( caContent.dwSubType ) {
      case SCCCA_DOCUMENTPROPERTY:
	if (EXTYPE_PROPERTIES == exType) {
	  AddDocumentProperty( hContent, caContent.dwData1, &caContent, outFile );
	}
	else {
	  SkipTagLocal( hContent, caContent.dwSubType, ( PSCCCAGETCONTENT ) & caContent );
	}
	break;

      case SCCCA_STYLE:
    //case SCCCA_FRAME:
    //case SCCCA_FRAME_EX:
    //case SCCCA_SUBDOCTEXT:
      case SCCCA_NAMEDCELLRANGE:
    //This  next tag, sccca_vectorsavetag, may contain text from an excel text box, so leave it in!
    //case SCCCA_VECTORSAVETAG: */
      case SCCCA_DATEDEFINITION:
      case SCCCA_CAPTIONTEXT:
      case SCCCA_COUNTERFORMAT:
      case SCCCA_CUSTOMDATAFORMAT:
      case SCCCA_INLINEDATAFORMAT:
	SkipTagLocal( hContent, caContent.dwSubType, ( PSCCCAGETCONTENT ) & caContent );
	break;

      default:
	break;
      }

      break;

    case SCCCA_ENDTAG:
      /*
	Linked objects handled on ENDTAG so any linked text appears
	before the expanded content.
      */
      if (caContent.dwSubType == SCCCA_LINKEDOBJECT) {
	if (nExpandAttachments != FALSE) {
	  IOSPECLINKEDOBJECT linkObj;
	  VTHDOC hLODoc = VTHDOC_INVALID;
	  
	  /*
	    Setup the linked object.
	  */
	  linkObj.dwStructSize = sizeof(IOSPECLINKEDOBJECT);
	  linkObj.hDoc = hLocDoc;
	  linkObj.dwObjectId = caContent.dwData1;
	  linkObj.dwType = caContent.dwData2;
	  linkObj.dwParam1 = caContent.dwData3;
	  linkObj.dwParam2 = caContent.dwData4;
	  linkObj.dwReserved1 = 0;
	  linkObj.dwReserved2 = 0;

	  /*  Use the locator to get an HDOC.  */
	  err = DAOpenDocument(&hLODoc, IOTYPE_LINKEDOBJECT, &linkObj, 0);

	  if (err == DAERR_OK) {
	    /*
	      +----------------------------------------------------------------------+
	      | Use the output character set specified by the global dwCharacterSet  |
	      | and pen the document for the content access.                         |
	      +----------------------------------------------------------------------+
	    */
	    err = DASetOption( hLODoc, SCCOPT_OUTPUTCHARACTERSET,
			       (VTLPVOID)&dwCharacterSet, sizeof(VTDWORD) );
	    if ( err == DAERR_OK ) {
	      if(hLODoc != VTHDOC_INVALID) {
		/*
		  Intentionally ignore errors in the sub-document in favor of
		  continuing processing.
		*/
		doCA(outFile, hLODoc);
	      }
	    }
	  }
	  
	  DACloseDocument(hLODoc);
	}
      }
      break;

    case SCCCA_OBJECT:
      if (nExpandEmbeddings != FALSE) {
	IOSPECSUBOBJECT sObject;
	VTHDOC hDocEmbedding;
 
	/*  Open a new DA and CA on the embedding.  */
	sObject.dwStructSize = sizeof(IOSPECSUBOBJECT);
	sObject.hDoc         = hLocDoc;
	sObject.dwObjectId   = caContent.dwData1;
	sObject.dwReserved1  = 0;
	sObject.dwReserved2  = 0;
	hDocEmbedding = VTHDOC_INVALID;

	/*  Try the OLE object if available.  */
	if (caContent.dwData3 != 0xFFFFFFFF) {
	  sObject.dwStreamId = caContent.dwData3;
	  err = DAOpenDocument(&hDocEmbedding, IOTYPE_SUBOBJECT, &sObject, 0);

	  /*
	    If we can't identify the object for the embedding, then
	    try the alternate graphic if available.
	  */
	  if (err == SCCERR_OK) {
	    VTDWORD dwFileId;
	    err = DAGetFileId(hDocEmbedding, &dwFileId);
	    if ((err != SCCERR_OK) || (dwFileId == FI_UNKNOWN)) {
	      DACloseDocument(hDocEmbedding);
	      hDocEmbedding = VTHDOC_INVALID;
	    }
	  }
	}

	/*  Now try other stream if first one failed.  */
	if ((hDocEmbedding == VTHDOC_INVALID) && (caContent.dwData2 != 0xFFFFFFFF)) {
	  /*  Try the alternate graphic.  */
	  sObject.dwStreamId = caContent.dwData2;
	  err = DAOpenDocument(&hDocEmbedding, IOTYPE_SUBOBJECT, &sObject, 0);
	}

	if (err == SCCERR_OK) {
	  if (hDocEmbedding != VTHDOC_INVALID) {
	    err = DASetOption(hDocEmbedding,
			      SCCOPT_OUTPUTCHARACTERSET,
			      (VTLPVOID)&dwCharacterSet,
			      sizeof(VTDWORD));
	    /*
	      Intentionally ignore errors in the sub-document in favor of
	      continuing processing.
	    */
	    doCA(outFile, hDocEmbedding);
	  }

	  DACloseDocument(hDocEmbedding);
	}
      }

      break;

    case SCCCA_FILEPROPERTY:
      break;

    case SCCCA_BREAK:
      break;
    }

    err = CAReadNext( hContent, &caContent );
  }

  /*
    Process documents stored in archive.  It relies upon the DAGetTreeCount()
    call to determine whether the document has compressed nodes that need to be
    expanded.  If so, each node is opened and then processed.  Errors returned
    during the processing of archive formats are not returned to the caller but
    are instead ignored in favor of attempting to continue processing of the
    remaining documents in the archive.
  */
  if (nExpandArchives != FALSE) {
    VTDWORD dwCount;
    if (DAGetTreeCount(hLocDoc, &dwCount) == DAERR_OK) {
      SCCDATREENODE sNode;

      sNode.dwSize = sizeof(SCCDATREENODE);
      for (sNode.dwNode = 0; sNode.dwNode < dwCount; sNode.dwNode++) {
	/*  Get the node.  */
	err = DAGetTreeRecord(hLocDoc, &sNode);
	if ((err == SCCERR_OK) && !(sNode.dwFlags & SCCDA_TREENODEFLAG_FOLDER)) {
	  /*  Not a folder, process.  */
	  VTHDOC hDocTreeRecord = VTHDOC_INVALID;
	  err = DAOpenTreeRecord(hLocDoc, &hDocTreeRecord, sNode.dwNode);
	  if (err == SCCERR_OK) {
	    err = DASetOption(hDocTreeRecord,
			      SCCOPT_OUTPUTCHARACTERSET,
			      (VTLPVOID)&dwCharacterSet,
			      sizeof(VTDWORD));
	    /*
	      Intentionally ignore errors in the sub-document in favor of
	      continuing processing.
	    */
	    doCA(outFile, hDocTreeRecord);
	  }
	  DACloseTreeRecord(hDocTreeRecord);
	}
      }
    }
  }

  /*
  ** Now that the source file has been processed
  ** free the buffer and clean up
  */
  free(hBuffer);
  CACloseContent(hContent);

  return err;
}


////////////////////////////////////////////////////////////////////////////////
void SkipTagLocal(
		  VTHCONTENT hContent,
		  VTDWORD dwTagType,
		  PSCCCAGETCONTENT pcaContent )
{
  DAERR daErr;
  int nNestLevel = 1;

  do {
    daErr = CAReadNext( hContent, pcaContent );
    if ( pcaContent->dwType == SCCCA_BEGINTAG ) {
      if ( pcaContent->dwSubType == dwTagType )
	nNestLevel++;
    } /* end_if */
    else if ( pcaContent->dwType == SCCCA_ENDTAG ) {
      if ( pcaContent->dwSubType == dwTagType )
	nNestLevel--;
    } /* end_if */
  }
  while ( ( daErr == DAERR_OK ) && ( nNestLevel > 0 ) );
}


////////////////////////////////////////////////////////////////////////////////
void AddDocumentProperty(
			 VTHCONTENT hContent,
			 VTDWORD dwPropertyId,
			 PSCCCAGETCONTENT pcaContent,
			 HANDLE* TempPropFile )
{
  DAERR   daErr;
  char *PropName = NULL;
  VTBOOL  bInsideName = FALSE;
  VTBOOL  bNotDone = TRUE;
  VTBOOL  bHaveText = FALSE;
  WORD w;
  DWORD dwBytesWritten;  

  /* Get the Property Name; if it can't be found, it might be user defined */
  if(dwPropertyId < SCCCA_USERDEFINEDPROP) {
    PropName = props[ dwPropertyId - 1 ];
  }

  // Property Name
  if (PropName != NULL) {
    unsigned int i;
    for(i=0; i<strlen(PropName); i++) {
      w = PropName[i];
      WRITE_FILE(TempPropFile, &w, sizeof(WORD), &dwBytesWritten);
    }
    w = ':';
    WRITE_FILE(TempPropFile, &w, sizeof(WORD), &dwBytesWritten);
    w = ' ';
    WRITE_FILE(TempPropFile, &w, sizeof(WORD), &dwBytesWritten);
  }

  /* Get all the text forming the value for this property */
  do {
    daErr = CAReadNext( hContent, pcaContent );

    bHaveText = FALSE;
    switch (pcaContent->dwType) {
    case SCCCA_BEGINTAG:
      if(( pcaContent->dwSubType == SCCCA_DOCUMENTPROPERTYNAME ) && ( dwPropertyId >= SCCCA_USERDEFINEDPROP ) && ( bInsideName == FALSE )) {
	bInsideName = TRUE;
      }
      break;
    case SCCCA_ENDTAG:
      if(( pcaContent->dwSubType == SCCCA_DOCUMENTPROPERTYNAME ) && ( dwPropertyId >= SCCCA_USERDEFINEDPROP ) && ( bInsideName == TRUE )) {
	bInsideName = FALSE;
      }
      else if( pcaContent->dwSubType == SCCCA_DOCUMENTPROPERTY ) {
	/*
	** When the final ENDTAG is found, the property value is complete
	*/
	bNotDone = FALSE;
      }
      break;
    case SCCCA_TEXT:
      bHaveText = TRUE;
      break;

    case SCCCA_GENERATED:
      bHaveText = TRUE;
      break;
    }

    /* Gather all the text we've found */
    if (bHaveText) {
      if(bInsideName) {
	// Property Name
	WRITE_FILE(TempPropFile, pcaContent->pDataBuf, pcaContent->dwDataBufSize, &dwBytesWritten);
	w = ':';
	WRITE_FILE(TempPropFile, &w, sizeof(WORD), &dwBytesWritten);
	w = ' ';
	WRITE_FILE(TempPropFile, &w, sizeof(WORD), &dwBytesWritten);
      }
      else {
	// Property Value
	WRITE_FILE(TempPropFile, pcaContent->pDataBuf, pcaContent->dwDataBufSize, &dwBytesWritten);
      }
    }
  }
  while ((daErr == DAERR_OK) && bNotDone);

  w = '\n';
  WRITE_FILE(TempPropFile, &w, sizeof(WORD), &dwBytesWritten);

  return;
}


////////////////////////////////////////////////////////////////////////////////
// output the special silent sentence marker
void WriteSeparator(HANDLE *hFile)
{
  DWORD dwBytesWritten;

  if(SO_UNICODE == dwCharacterSet) {
    WORD w = ' ';
    WRITE_FILE(hFile, &w, sizeof(WORD), &dwBytesWritten);
    w = '\x1E';
    WRITE_FILE(hFile, &w, sizeof(WORD), &dwBytesWritten);
    w = ' ';
    WRITE_FILE(hFile, &w, sizeof(WORD), &dwBytesWritten);
  }
  else {
    WRITE_FILE(hFile, " \x1E ", 3, &dwBytesWritten);
  }
}
